package net.chlup.myshare;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.nio.charset.Charset;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.TimeZone;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.entity.mime.HttpMultipartMode;
import org.apache.http.entity.mime.MultipartEntity;
import org.apache.http.entity.mime.content.InputStreamBody;
import org.apache.http.entity.mime.content.StringBody;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.client.DefaultHttpRequestRetryHandler;
import org.apache.http.util.EntityUtils;

import android.app.IntentService;
import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.content.ContentResolver;
import android.content.Context;
import android.content.Intent;
import android.database.Cursor;
import android.net.Uri;
import android.os.Bundle;
import android.os.Environment;
import android.os.PowerManager;
import android.provider.MediaStore;
import android.util.Base64;
import android.util.Log;

public class myshareservice extends IntentService {

   /* Prefs */
   Preferences prefs;
   DefaultHttpClient httpclient = null;

   /* Storage PATH */
   String storagepath = Environment.getExternalStorageDirectory().toString();

   /* Notification */
   NotificationManager notman;
   PendingIntent notintent;
   
   int sendOK;
   int sendError;

   public myshareservice() {
      super("myshareservice");
   }

   @Override
   protected void onHandleIntent(Intent intent) {

      /* Get prefs */
      this.prefs = new Preferences(this);
      this.prefs.load();

      setupHttpConnection();

      /* nedovol, aby usnulo CPU behem  uploadu (pokud to sem nedam a usnuti dovolim, fungovalo to pekne, ale pri usnuti telefonu se upload prerusil a bezel az po jeho probuzeni) */
      PowerManager pm = (PowerManager) getSystemService(Context.POWER_SERVICE);
      PowerManager.WakeLock wl = pm.newWakeLock(PowerManager.PARTIAL_WAKE_LOCK, "myshareserviceStart - CPU no sleep");
      wl.acquire();
         myshareserviceStart(intent);
      wl.release();

   }

   private void setupHttpConnection() {
      this.httpclient = new DefaultHttpClient();
      this.httpclient.setHttpRequestRetryHandler(new DefaultHttpRequestRetryHandler(1, false)); // retry jen 1x (napriklad je nedostupna sit) (pokud bych to neuvedl, tak defaultne je tusim retry 3x)
      if(this.prefs.own_key_store) {
         try {
            KeyStore trustStore = KeyStore.getInstance(KeyStore.getDefaultType());
            FileInputStream instream = new FileInputStream(new File(storagepath + "/security/cacerts.bks"));
            trustStore.load(instream, "changeit".toCharArray());
            instream.close();
            SSLSocketFactory socketFactory = new SSLSocketFactory(trustStore);
            Scheme sch = new Scheme("https", socketFactory, 443); // deprecated
            //Scheme sch = new Scheme("https", 443, socketFactory); // takhle to ma byt nove, ale hrouti se to (prej: java.lang.NoSuchMethodError) - jako by to davalo prednost verzi co je v Androidu a ne te co jsem tam vsunul pres externi knihovny - tak zustavam pri stare verzi 
            this.httpclient.getConnectionManager().getSchemeRegistry().register(sch);
         }
         catch (KeyStoreException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (FileNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (NoSuchAlgorithmException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (CertificateException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (KeyManagementException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (UnrecoverableKeyException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
      }
   }

   private void setupNotif() {
      /* Show state in notif bar */

      if (this.notman != null) {
         /* Remove notification */
         this.notman.cancelAll();
         this.notman = null;
      }

      /* Set up a persistent notification */
      this.notman = (NotificationManager)   getSystemService(Context.NOTIFICATION_SERVICE);
      this.notman.cancelAll();
      this.notintent = PendingIntent.getActivity(
            this,
            0,
            new Intent(this, myshare.class),
            0
      );
   }

   private void myshareserviceStart(Intent intent) {

      Bundle extras = intent.getExtras();
      String action = intent.getAction();
      //String type = intent.getType(); // obsahuje mimetype
      int serverIndex = extras.getInt("MYSHARE_SERVERINDEX");

      if(this.prefs.target_urls.get(serverIndex) != null) {
         String result_last = "Error: unknown";
         this.sendOK=0;
         this.sendError=0;
         setupNotif();
         setNotif(0, 0, myshareservice.this.getString(R.string.msgUploadStart), 0);
         // if this is from the share menu
         if (Intent.ACTION_SEND.equals(action)) {
            if (extras.containsKey(Intent.EXTRA_STREAM)) {
               Log.i(this.getClass().getName(), "upload one file");
               // Get resource path from intent callee
               Uri uri = (Uri) extras.getParcelable(Intent.EXTRA_STREAM);
               String result = mainSendRequest(intent, uri, serverIndex, extras.getStringArray("MYSHARE_TAGS"), false);
               setNotif(0, 0, result, 1);
               result_last = result;
            }
            else if (extras.containsKey(Intent.EXTRA_TEXT)) {
               Log.i(this.getClass().getName(), "upload one text");
               String result = mainSendRequestText(intent, serverIndex, extras.getStringArray("MYSHARE_TAGS"));
               setNotif(0, 0, result, 1);
               result_last = result;
            }
         }
         // if this is from the share menu
         else if (Intent.ACTION_SEND_MULTIPLE.equals(action)) {
            if (extras.containsKey(Intent.EXTRA_STREAM)) {
               Log.i(this.getClass().getName(), "upload list files");
               ArrayList<Uri> theFileList = intent.getParcelableArrayListExtra(Intent.EXTRA_STREAM);
               int id = 0;
               for(Uri uri: theFileList) {
                  String result = mainSendRequest(intent, uri, serverIndex, extras.getStringArray("MYSHARE_TAGS"), true);
                  setNotif(id, theFileList.size()-1, result, 1);
                  result_last = result;
                  ++id;
               }
            }
         }
         setNotif(0, 0, result_last, 2);
      }

   }

   private String mainSendRequestText(Intent intent, int serverIndex, String[] tags) {
      String result = "Error: unknown";
      InputStream is;
      String sharedText = "";
      if(intent.getStringExtra(Intent.EXTRA_TEXT) != null) {
         sharedText = intent.getStringExtra(Intent.EXTRA_TEXT);
      }
      try {
         is = new ByteArrayInputStream(sharedText.getBytes("UTF-8"));
         String filenameshort = "unknown";
         if(intent.getStringExtra(Intent.EXTRA_SUBJECT) != null) {
            filenameshort = intent.getStringExtra(Intent.EXTRA_SUBJECT);
         }
         String createdate = "";
         String corectiontogmt = "";
         // samotne odeslani dat
         result = SendRequest(is, filenameshort, createdate, corectiontogmt, serverIndex, tags);
     }
     catch (UnsupportedEncodingException e) {
         result = "Error: " + myshareservice.this.getString(R.string.msgUploadFailed) + " (text)";
         Log.e(this.getClass().getName(), e.toString());
      }
      return result;
   }
   
   private String mainSendRequest(Intent intent, Uri uri, int serverIndex, String[] tags, Boolean multiple) {
      String result = "Error: unknown";
      // Query gallery for camera picture via
      // Android ContentResolver interface
      ContentResolver cr = getContentResolver();
      InputStream is;
      try {
         // zjisteni InputStream obrazku dle uri 
         is = cr.openInputStream(uri);
         // zjisteni, kdy byl soubor naposledy modifikovan ("vytvoren")
         String filenameshort = "unknown";
         String createdate = "";
         String corectiontogmt = "";
         String filename = getRealPathFromContentUri(uri);
         //Log.i(this.getClass().getName(), "uri: " + uri);
         //Log.i(this.getClass().getName(), "filename: " + filename);
         if(filename != null) {
            File file = new File(filename);
            if(file.exists()) {
               filenameshort = file.getName();
               long epochlastmodified=file.lastModified();
               //createdate = new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss").format(new java.util.Date (epochlastmodified));
               SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
               sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
               createdate = sdf.format(new Date(epochlastmodified));
               corectiontogmt = Integer.toString(0); // cas posilam v GMT a proto je korekce na GMT vzdy 0
               Log.i(this.getClass().getName(), "createdate: " + createdate);
            }
         }
         if(!multiple) {
            // davam prednost jmena z EXTRA_SUBJECT pred skutecnym jmenem souboru
            // (je to dobre pri posilani obrazku ci zvuku z aplikace Note Everything)
            if(intent.getStringExtra(Intent.EXTRA_SUBJECT) != null) {
               filenameshort = intent.getStringExtra(Intent.EXTRA_SUBJECT);
            }
         }
         // samotne odeslani dat
         result = SendRequest(is, filenameshort, createdate, corectiontogmt, serverIndex, tags);
      }
      catch (FileNotFoundException e) {
         result = "Error: " + myshareservice.this.getString(R.string.msgUploadFailed) + " (file)";
         Log.e(this.getClass().getName(), e.toString());
      }
      // Get binary bytes for encode
      return result;
   }

   private String SendRequest(InputStream is, String filenameshort, String createdate, String corectiontogmt, int serverIndex, String[] tags) {

      String result = "Error: unknown";

      /* reseni problemu s nepodarenou inicializaci httpclient
       *  - vetsinou pujde o chybu, ze nejsou dobre nacteny vlastni CA
       *    z lokalniho souboru */
      if(httpclient == null) {
         Log.e(this.getClass().getName(), "httpclient is null (may be problems with CA)");
         result = "Error: httpclient is null (may be problems with CA)";
         return result;
      }

      URL target_url = this.prefs.target_urls.get(serverIndex);
      String http_user = this.prefs.http_users.get(serverIndex);
      String http_password = this.prefs.http_passwords.get(serverIndex);
      Log.w(this.getClass().getName(), "http_user: "+http_user);
      HttpPost httppost = new HttpPost(target_url.toString());
      if (http_user != null && http_password != null) {
         String http_userpassword = http_user + ":" + http_password;
         /* verze i pro Android 2.1:
          * vyzaduje:
          *    externi knihovnu: commons-codec-1.4.jar
          *    import org.apache.commons.codec.binary.Base64;
          *    import org.apache.commons.codec.binary.StringUtils;
          * a NESMI byt pritomno:
          *    import android.util.Base64;
          */
         //httppost.setHeader(
         //      "Authorization",
         //      "basic " + new String(Base64.encodeBase64(StringUtils.getBytesUtf8(http_userpassword)))
         //);
         /* verze vyzadujici Android 2.2 a vyssi:
          * vyzaduje:
          *    import android.util.Base64;
          * a NESMI byt pritomno:
          *    import org.apache.commons.codec.binary.Base64;
          */
         httppost.setHeader("Authorization", "basic " + Base64.encodeToString(http_userpassword.getBytes(),Base64.NO_WRAP));
      }

      HttpResponse response;

      try {
         String serverinfo="Error: no message from the server";

         //MultipartEntity mpEntity = new MultipartEntity();
         MultipartEntity mpEntity = new MultipartEntity(HttpMultipartMode.BROWSER_COMPATIBLE, null, Charset.forName("UTF-8")); // diky uvedenym parametrum se pekne prenese nazev souboru obsahujici diakritiku
         mpEntity.addPart("img", new InputStreamBody(is, filenameshort)); // take by se zdal uvest mimetyp (pred filenameshort)
         mpEntity.addPart("createdate", new StringBody(createdate));
         mpEntity.addPart("corectiontogmt", new StringBody(corectiontogmt));
         if(this.prefs.prefer_exif_date) {
            mpEntity.addPart("preferexifdate", new StringBody("1"));
         }
         else {
            mpEntity.addPart("preferexifdate", new StringBody("0"));
         }
         for(int tagnum=0; tagnum<tags.length; ++tagnum) {
            mpEntity.addPart("tag", new StringBody(tags[tagnum], Charset.forName("UTF-8")));
         }

         httppost.setEntity(mpEntity);
         response = httpclient.execute(httppost);
         HttpEntity entity = response.getEntity();
         if (entity != null) {
            serverinfo=EntityUtils.toString(entity); // tohle muze asi pripadne nahradit radek nize (pokud budes chti s obsahem co je vracen pracovat)
            if(serverinfo.length()>100) {
               serverinfo=serverinfo.substring(0,100);
            }
            //entity.consumeContent(); // zajisti asi konzumaci obsahu s ukoncenim pozadavku - diky tomu funguje keep-alive (kdyz to tu nebylo, keep-alive nefungoval - nestacilo jen posunout definic httpclient nahoru)
         }
         result = serverinfo;
      }
      catch (ClientProtocolException e) {
         Log.e(this.getClass().getName(), "ClientProtocolException: " + e.toString());
         result = "Error: ClientProtocolException";
      }
      catch (IOException e) {
         Log.e(this.getClass().getName(), "IOException: " + e.toString());
         result = "Error: " + myshareservice.this.getString(R.string.msgUploadFailed);
      }
      return result;
   }

   private void setNotif(int id, int maxid, String text, int type) {
      int idnotif = 0;
      int icon = R.drawable.notif_icon;
      int flag = Notification.FLAG_ONGOING_EVENT;
      if(type==1) {
         if(text.length()<2 || text.substring(0,2).compareTo("OK")!=0) {
            ++this.sendError;
            // pridavnou Error chybu v notify prihazuji jen pri odesilani vice jak jednoho dokumentu
            if(maxid!=0) { 
               setNotif(id, maxid, text, -1);
            }
         }
         else {
            ++this.sendOK;
         }
         if(maxid!=0) {
            text = Integer.toString(id+1) + "/" + Integer.toString(maxid+1) + " " + text;
         }
      }
      else if(type==-1) { 
         idnotif = id+1;
         icon = R.drawable.notif_icon_failed;
         flag = Notification.FLAG_AUTO_CANCEL;      
         if(maxid!=0) {
            text = Integer.toString(id+1) + "/" + Integer.toString(maxid+1) + " " + text;
         }
      }
      else if(type==2) {
         // neni chyba a byl odeslan jeden ci vice dokumentu (ve finale je tedy jen jedna notify zprava - tahle)
         if(this.sendError==0 && this.sendOK>0) {
            icon = R.drawable.notif_icon_success;
         }
         // je chyba a byl odesilan jen jeden dokument (ve finale je tedy jen jedna notify zprava - tahle)
         else if(this.sendError>0 && (this.sendOK+this.sendError)==1) {
            icon = R.drawable.notif_icon_failed;
         }
         // dokumentu bylo odesilano vice -> zprava z posledniho vysledku
         // bude nahrazena zpravou souhrnneho charakteru
         if((this.sendOK+this.sendError)>1) {
            text = myshareservice.this.getString(R.string.msgUploadEnd);
            text += " (";
            if(this.sendError>1) {
               text += "Error:" + Integer.toString(this.sendError) + "x, ";
            }
            text += "OK:" + Integer.toString(this.sendOK) + "x";
            text += ")";
         }
         flag = Notification.FLAG_AUTO_CANCEL;      
      }
      Notification notif = new Notification(
            icon,
            null, // neuvadim zadny text, aby zbytecne na status baru nebyly sachry
            System.currentTimeMillis()
      );
      notif.flags = flag;
      notif.setLatestEventInfo(
            this,
            getString(R.string.app_name),
            text,
            this.notintent
      );
      this.notman.notify(idnotif, notif);
   }

   // Convert the image/file URI to the direct file system path of the file
   public String getRealPathFromContentUri(Uri contentUri) {
      String filePath = null;
      if(contentUri.getScheme().compareTo("content")==0) {
         String [] filePathColumn = { MediaStore.Images.Media.DATA };
         Cursor cursor = getContentResolver().query(contentUri, filePathColumn, null, null, null);
         if(cursor != null) {
            cursor.moveToFirst();
            int columnIndex = cursor.getColumnIndex(filePathColumn[0]);
            filePath = cursor.getString(columnIndex);
            cursor.close();
         }
         //int column_index = cursor.getColumnIndexOrThrow(MediaStore.Images.Media.DATA);
      }
      else if(contentUri.getScheme().compareTo("file")==0) {
         filePath = contentUri.getPath();
      }
      return filePath;
   }

}
